#!/usr/bin/python
# Copyright (c) 2003-2013  Pavel Rychly, Vojtech Kovar, Milos Jakubicek, Milos Husak, Vit Baisa

import manatee
import os.path, sys, glob
from types import UnicodeType
try:
    from hashlib import md5
except ImportError:
    from md5 import new as md5
from butils import *

class CorpusManager:
    def __init__ (self, corplist=['susanne'], subcpath=[], gdexpath=[]):
        self.corplist = corplist
        self.subcpath = subcpath
        self.gdexdict = dict(gdexpath)

    def default_subcpath(self, corp):
        if type(corp) is UnicodeType: corp = self.get_Corpus (corp)
        if corp.get_conf('SUBCBASE'): return corp.get_conf('SUBCBASE')
        cpath = corp.get_conf('PATH')
        return os.path.join(cpath, 'subcorp')

    def get_Corpus (self, corpname, subcname=''):
        if ':' in corpname:
            corpname, subcname = corpname.split(':',1)
        corp = manatee.Corpus (corpname)
        manatee.setEncoding(corp.get_conf('ENCODING'))
        corp.corpname = str(corpname) # never unicode (paths)
        corp.cm = self
        dsubcpath = self.default_subcpath (corp)
        if subcname:
            for sp in self.subcpath + [dsubcpath]:
                if sp == dsubcpath:
                    spath = os.path.join (sp, subcname + '.subc')
                else:
                    spath = os.path.join (sp, corpname, subcname + '.subc')
                if type(spath) == unicode:
                    spath = spath.encode("utf-8")
                if os.path.isfile (spath):
                    subc = manatee.SubCorpus (corp, spath)
                    subc.corp = corp
                    subc.spath = spath
                    try: open(spath[:-4] + 'used', 'w')
                    except: pass
                    subc.corpname = str(corpname) # never unicode (paths)
                    subc.subcname = subcname
                    subc.cm = self
                    subc.subchash = md5(open(spath).read()).digest()
                    return subc
            raise RuntimeError (_('Subcorpus "%s" not found') % subcname)
        else:
            return corp

    def findPosAttr (self, corpname, attrname):
        return manatee.findPosAttr (corpname.split(':',1)[0], attrname)
            
    def corpconf_pairs (self, corp, label):
        if type(corp) is UnicodeType:
            corp = self.get_Corpus (corp)
        val = corp.get_conf(label)
        if len(val) > 2: val = val[1:].split(val[0])
        else: val = ''
        return [val[i:i+2] for i in range (0, len(val), 2)]

    def corpattrlist (self, corp):
        return dict (self.corpconf_pairs (corp, label)).get (item)

    def corplist_with_names (self, conf_options=[]):
        """conf_options 
           -- list of options to be extracted from corpus configure files"""
        #subc = manatee.StrVector()
        #for s in self.subcpath:
        #    manatee.find_subcorpora (s, subc)
        #corpora = self.corplist + map (None, subc)
        #corpora.sort()
        cl = []
        for c in self.corplist:
            try:
                mc = manatee.Corpus(c)
                cinf = {'id': c,
                        'name': mc.get_conf('NAME') or c}
                cinf.update([(o.lower(), mc.get_conf(o)) for o in conf_options])
            except:
                cinf = {'id': c, 'name': c}
            cl.append (cinf)
        cl.sort()
        return cl

    def subcorpora (self, corpname):
        # we must encode for glob.glob otherwise it fails for non-ascii files
        enc_corpname = corpname.encode("utf-8")
        subc = []
        for sp in self.subcpath:
            subc += glob.glob (os.path.join (sp, enc_corpname, '*.subc'))
        subc += glob.glob (os.path.join (
                self.default_subcpath (corpname).encode("utf-8"), '*.subc'))
        return sorted(subc)

    def subcorp_names (self, corpname):
        return [{'n': os.path.splitext(os.path.basename (s))[0]}
                for s in self.subcorpora (corpname)]

    def last_subcorp_names (self, corpname, maxitems=25):
        
        subc = [os.path.isfile(c[:-4]+'used') and c[:-4]+'used' or c
                for c in self.subcorpora (corpname)]
        subc = [(os.stat(c).st_ctime, c) for c in subc]
        subc.sort(reverse=True)
        return [{'n': os.path.splitext(os.path.basename (s))[0]}
                for t,s in subc[:maxitems]]

    def find_same_subcorp_file (self, subcorp, wanted_infix,
                                 wanted_suff=('arf','frq', 'docf')):
        basedir = os.path.dirname (self.subcpath[0].rstrip('/'))
        scpath = subcorp.spath
        scsize = os.path.getsize (scpath)
        scdata = open(scpath).read()

        for f in glob.glob (basedir + '/*/%s/*.subc' % subcorp.corpname):
            if f == scpath: continue
            if os.path.getsize(f) != scsize: continue
            if [s for s in wanted_suff
                if not os.path.isfile(f[:-4] + wanted_infix + '.' + s)]:
                continue
            # now, f has same size and all wanted suffixes exist
            if open(f).read() == scdata:
                return f
        return None


def ws_subc_freq (wmap3, corp):
    if hasattr(corp, 'spath'):
        fs = wmap3.poss()
        fs2 = corp.filter_fstream(fs)
        return corp.count_rest(fs2)
    return wmap3.getcnt()

def ws_find_triple (wmap1, id1, id2, id3):
    # WARNING: ids has to be bigger than the current position in wmap
    if not wmap1.findid (id1): return None
    wmap2 = wmap1.nextlevel()
    if not wmap2.findid (id2): return None
    wmap3 = wmap2.nextlevel()
    if not wmap3.findid (id3): return None
    return wmap3


def ngrams(corp, words=[], wlattr='', wlpat='', wlminfreq=5, wlmaxitems=100,
          wlsort='', blacklist=[], include_nonwords=0, wlmaxfreq=0, ngrams_n=2):
    import re
    from operator import itemgetter
    ngr_data_filename = str(corp.get_conf('PATH') + wlattr + '.ngr')
    try: ngr = manatee.NGram(ngr_data_filename)
    except: raise MissingSubCorpFreqFile(ngr_data_filename)
    attr = corp.get_attr(wlattr)
    if hasattr(corp, 'spath'):
        freqs = frq_db (corp, wlattr+'.ngr', id_range=ngr.size())
    items = []
    words, blacklist = set(words), set(blacklist)
    nwre = re.compile(corp.get_conf('NONWORDRE').replace('^[:alpha:]', '\d\W_'),
                      re.UNICODE)
    if wlpat and wlpat != '.*': wlpatre = re.compile(wlpat, re.UNICODE)
    else: wlpatre = None
    for i in xrange(ngr.size()):
        if hasattr(corp, 'spath'): freq = freqs[i]
        else: freq = ngr.freq(i)
        if freq < wlminfreq or (wlmaxfreq and freq > wlmaxfreq): continue
        if ngrams_n < ngr.min(i): continue # n-gram too big
        ngrwords = []
        lids = ngr.lids(i)
        whitelisted = False
        for n in xrange(ngr.max(i)): # "unzip" n-gram and filter
            word = attr.id2str(lids.next())
            if not include_nonwords and nwre.match(word): break
            if blacklist and word in blacklist: break
            if words and word in words: whitelisted = True
            ngrwords.append(word)
            if n+1 == ngrams_n: # check and append
                ngram = ' '.join(ngrwords)
                if wlpatre and not wlpatre.search(ngram): continue
                if words and not whitelisted and not ngram in words: continue
                if blacklist and ngram in blacklist: continue
                items.append((freq, ngram))
        if len (items) > 5 * wlmaxitems:
            if wlsort == 'f': items.sort(reverse=True)
            else: items.sort(key=itemgetter(1))
            del items [wlmaxitems:]
    if wlsort == 'f': items.sort(reverse=True)
    else: items.sort(key=itemgetter(1))
    del items [wlmaxitems:]
    return [{'str': w, 'freq': f} for f, w in items]


def ws_wordlist (corp, wlmaxitems=100, wlsort='', wlattr='ws_collocations'):
    import wmap
    wmap.setEncoding('utf-8')
    wsbasekw = 'TERMBASE' if wlattr == 'ws_terms' else 'WSBASE'
    csize = corp.search_size()
    basewsbase = os.path.basename (corp.get_conf(wsbasekw))
    freqs_file = subcorp_base_file (corp, basewsbase + '.hfrq')
    lex_file = subcorp_base_file (corp, basewsbase + '.hlex')
    logfile = wlattr + '.build'
    if not os.path.isfile (freqs_file): # not computed
        raise MissingSubCorpFreqFile (corp)
    if os.path.isfile (subcorp_base_file(corp, logfile)):
        raise MissingSubCorpFreqFile (corp) # computation in progress

    result_str = wmap.StrVector()
    wmap.terms_lexicon (freqs_file, lex_file, result_str, wlmaxitems)

    try: wsstrip = int(corp.get_conf('WSSTRIP'))
    except ValueError: wsstrip = 0
    # find out seek and cnt
    ws1 = wmap.WMap (corp.get_conf(wsbasekw), 0, 0, 0, corp.get_conffile())
    result_parsed = []
    for item in result_str:
        w1, gramrel, w2 = item.split ('\t')[:3]
        result_parsed.append((ws1.coll2id(w1), ws1.str2id(gramrel),
                              ws1.coll2id(w2), w1, gramrel, w2))
    result = []
    for id1, id2, id3, w1, gramrel, w2 in sorted (result_parsed):
        ws3 = ws_find_triple (ws1, id1, id2, id3)
        if not ws3: continue
        freq = ws_subc_freq (ws3, corp)
        w2s = wsstrip and w2[:-wsstrip] or w2
        result.append ((-freq, w1, gramrel, w2s, ws3.tell()))
    return [{'w1': w1, 'gramrel': g, 'w2': w2, 'seek': seek,
            'freq': -mfreq, 'str': ' '.join([w1, g, w2]) }
                             for mfreq, w1, g, w2, seek in sorted (result)]


def wordlist (corp, words=[], wlattr='', wlpat='', wlminfreq=5, wlmaxitems=100,
              wlsort='', blacklist=[], wlnums='frq', include_nonwords=0,
              wlmaxfreq=0):
    blacklist = set (blacklist)
    words = set (words)
    attr = corp.get_attr (wlattr)
    if '.' in wlattr: # attribute of a structure
        struct = corp.get_struct(wlattr.split('.')[0])
        if wlnums == 'doc sizes':
            normvals = dict ([(struct.beg(i), struct.end(i)-struct.beg(i))
                                              for i in xrange(struct.size())])
        else:
            normvals = dict ([(struct.beg(i),1) for i in xrange(struct.size())])
        attrfreq = dict ([(i, doc_sizes(corp, struct, wlattr, i, normvals))
                                              for i in xrange(attr.id_range())])
    else: # positional attribute
        attrfreq = frq_db (corp, wlattr, wlnums)
    items = []
    if words and wlpat == '.*': # word list just for given words
        for word in words:
            if wlsort == 'f':
                if len (items) > 5 * wlmaxitems:
                    items.sort()
                    del items [:-wlmaxitems]
            else:
                if len (items) >= wlmaxitems:
                    break
            id = attr.str2id(word)
            if id == -1:
                frq = 0
            else:
                frq = attrfreq[id]
            if word and frq >= wlminfreq \
                    and (not wlmaxfreq or frq <= wlmaxfreq) \
                    and (not blacklist or word not in blacklist):
                if wlnums == 'arf':
                    items.append ((round(frq, 1), word))
                else:
                    items.append ((frq, word))
    else: # word list according to pattern
        if not include_nonwords: nwre = corp.get_conf('NONWORDRE')
        else: nwre = ''
        try: gen = attr.regexp2ids (wlpat.strip(), 0, nwre)
        except TypeError: gen = attr.regexp2ids (wlpat.strip(), 0)
        while not gen.end():
            if wlsort == 'f':
                if len (items) > 5 * wlmaxitems:
                    items.sort()
                    del items [:-wlmaxitems]
            else:
                if len (items) >= wlmaxitems:
                    break
            id = gen.next()
            frq = attrfreq [id]
            if not frq: continue
            if frq >= wlminfreq and (not wlmaxfreq or frq <= wlmaxfreq) \
                    and (not words or attr.id2str(id) in words) \
                    and (not blacklist or attr.id2str(id) not in blacklist):
                if wlnums == 'arf':
                    items.append ((round(frq, 1), id))
                else:
                    items.append ((frq, id))
    if wlsort == 'f':
        items.sort()
        del items [:-wlmaxitems]
        items.reverse()
    if not words or wlpat != '.*':
        items = [(f,attr.id2str(i)) for (f,i) in items]
    if not wlsort:
        items.sort(key=lambda x: x[1].lower())
    return [{'str': w, 'freq': f} for f,w in items]

def doc_sizes(corp, struct, attrname, i, normvals):
    r = corp.filter_query (struct.attr_val (attrname.split('.')[1], i))
    cnt = 0
    while not r.end():
        cnt += normvals[r.peek_beg()]
        r.next()
    return cnt

def attr_vals (corpname, avattr, avpattern, avmaxitems=20):
    attr = manatee.findPosAttr (corpname, avattr)
    gen = attr.regexp2ids ('.*%s.*' % avpattern.strip(), True)
    items = []
    while not gen.end() and avmaxitems > 0:
        items.append (attr.id2str(gen.next()))
        avmaxitems -= 1
    if not items:
        return {'query': avpattern, 'suggestions': [_('--nothing found--')]}
    return {'query': avpattern,
            'suggestions': [','.join(["'" + item + "'" for item in items])]}

def texttype_values (corp, subcorpattrs, list_all=False, hidenone=True):
    if subcorpattrs == '#': return []
    attrlines = []
    for subcorpline in subcorpattrs.split(','):
        attrvals = []
        for n in subcorpline.split('|'):
            if n in ('', '#'):
                continue
            attr = corp.get_attr (n)
            attrval = { 'name': n,
                        'label': corp.get_conf (n+'.LABEL') or n,
                        'attr_doc': corp.get_conf (n+'.ATTRDOC'),
                        'attr_doc_label': corp.get_conf (n+'.ATTRDOCLABEL'),
                      }
            try:
                maxlistsize = int(corp.get_conf (n+'.MAXLISTSIZE'))
            except ValueError:
                maxlistsize = 22
            hsep = corp.get_conf(n+'.HIERARCHICAL')
            multisep = corp.get_conf(n+'.MULTISEP')
            if not hsep and not list_all \
                                and (corp.get_conf (n+'.TEXTBOXLENGTH')
                                      or attr.id_range() > maxlistsize):
                attrval ['textboxlength'] = (corp.get_conf (n+'.TEXTBOXLENGTH')
                                             or 24)
            else: # list of values
                if corp.get_conf(n+'.NUMERIC'):
                    vals = []
                    for i in xrange(attr.id_range()):
                        try: vals.append({'v': int(attr.id2str(i))})
                        except: vals.append({'v': attr.id2str(i)})
                elif hsep: # hierarchical
                    vals = [{'v': attr.id2str(i)}
                                  for i in xrange(attr.id_range())
                                  if not multisep in attr.id2str(i)]
                else:
                    vals = [{'v': attr.id2str(i)}
                            for i in xrange(attr.id_range())
                            if not (multisep and multisep in attr.id2str(i))]
                if hidenone: # hide empty + ===NONE===
                    i = 0
                    while i < len(vals):
                        if vals[i]['v'] in ('', '===NONE==='): del vals[i]
                        else: i += 1
                if hsep: # hierarchical
                    attrval ['hierarchical'] = hsep
                    attrval ['Values'] = get_attr_hierarchy(vals, hsep, multisep)
                else:
                    vals.sort()
                    attrval ['Values'] = vals
            attrvals.append (attrval)
        attrlines.append ({'Line': attrvals})
    return attrlines

def get_attr_hierarchy(vals, hsep, multisep):
    result = {}
    values = set([])
    for v in vals:
        values.add(v['v'])
    for value in sorted(values):
        level = result
        while hsep in value:
            key, value = value.split(hsep, 1)
            try: level = level[key]
            except: value = key + hsep + value; break
        level[value] = {}
    return print_attr_hierarchy(result, hsep=hsep)

def print_attr_hierarchy(layer, level=0, label='', hsep='::'):
    if not layer: return []
    result = []
    if level > 0: startdiv = True
    else: startdiv = False
    for item in sorted(layer):
        sub = print_attr_hierarchy(layer[item], level+1, label+hsep+item, hsep)
        if sub: display_plus = True
        else: display_plus = False
        if label: full_value = label[len(hsep):] + hsep + item
        else: full_value = item
        result.append({ 'v': full_value,
                        'key': label,
                        'label': item,
                        'shift': level * 16,
                        'startdiv': startdiv,
                        'enddiv': 0,
                        'display_plus': display_plus,
                      })
        startdiv = False
        result.extend(sub)
    if level > 0: result[-1]['enddiv'] += 1
    return result

def subc_freqs (subcorp, attr, minfreq=50, maxfreq=10000, last_id=None):
    return [(i, subcorp.count_rest(attr.id2poss(i)))
            for i in xrange(last_id or attr.id_range())
            if maxfreq > attr.freq(i) > minfreq]

def subc_keywords1 (subcorp, attr, minfreq=50, maxfreq=10000):
    p = (subcorp.size() - subcorp.search_size()) / float(subcorp.search_size())
    freqs = [(float(f) / (attr.freq(i) - f +1) * p, f, i)
             for (i, f) in subc_freqs (subcorp, attr, minfreq, maxfreq,
                                      attr.id_range()/1000)]
    #freqs.sort()
    #del freqs[:-maxitems]
    return freqs
    
def subc_keywords (subcorp, attr, minfreq=50, maxfreq=10000, last_id=10000,
                   maxitems=100):
    p = (subcorp.size() - subcorp.search_size()) / float(subcorp.search_size())
    candidates = []
    for i in xrange(last_id or attr.id_range()):
        if not (maxfreq > attr.freq(i) > minfreq):
            continue
        freq = subcorp.count_rest (attr.id2poss(i))
        if freq < 3:
            continue
        arf = subcorp.count_ARF (attr.id2poss(i), freq)
        score = arf / (attr.freq(i) - arf +1) * p
        #if score < 2.0:
        #    continue
        candidates.append ((score, arf, freq, i))
    candidates.sort()
    del candidates[:-maxitems]
    return candidates

def subcorp_base_file (corp, attrname):
    if hasattr(corp, 'spath'):
        return corp.spath[:-4] + attrname
    else:
        return corp.get_conf('PATH') + attrname
    
class MissingSubCorpFreqFile (Exception):
    pass

def frq_db (corp, attrname, nums='frq', id_range=0):
    import array, exceptions
    filename = subcorp_base_file (corp, attrname) + '.' + nums
    if not id_range: id_range = corp.get_attr (attrname).id_range()
    if nums == 'arf':
        frq = array.array('f')
        try:
            frq.fromfile (open(filename), id_range)
        except IOError:
            raise MissingSubCorpFreqFile (corp)
        except exceptions.EOFError:
            os.remove(filename.rsplit('.', 1)[0]+'.docf')
            raise MissingSubCorpFreqFile (corp)
    else:
        try:
            if corp.get_conf ('VIRTUAL') and not hasattr(corp, 'spath') \
                                         and nums == 'frq':
                raise IOError
            frq = array.array('i')
            frq.fromfile (open(filename), id_range)
        except exceptions.EOFError:
            os.remove(filename.rsplit('.', 1)[0]+'.docf')
            raise MissingSubCorpFreqFile (corp)
        except IOError:
            try:
                frq = array.array('l')
                frq.fromfile (open(filename + '64'), id_range)
            except IOError:
                if not hasattr(corp, 'spath') and nums == 'frq':
                    a = corp.get_attr (attrname)
                    frq.fromlist ([a.freq(i) for i in xrange(a.id_range())])
                else:
                    raise MissingSubCorpFreqFile (corp)
    return frq


def subc_keywords3 (sc, scref, attrname, minarf=10, maxitems=100):
    f = frq_db (sc, attrname)
    fref = frq_db (scref, attrname)
    p = sum(fref) / sum(f)
    attr = sc.get_attr(attrname)
    items = [(p * (f[i]+1) / (fref[i]+1), i)
             for i in xrange (attr.id_range())
             if f[i] >= minarf and (fref[i] == 0 or p * f[i] / fref[i] > 2.0)
             ]
    items.sort(reverse=True)
    del items[maxitems:]
    fsf = frq_db (sc, attrname)
    fcf = frq_db (scref, attrname)
    return [(s, attr.id2str(i), f[i], fref[i], fsf[i], fcf[i], i)
            for s,i in items]

def subc_keywords_onstr (sc, scref, wlattr='word', wlminfreq=5, wlpat='.*',
                         wlmaxitems=100, simple_n=100, words=[], wlmaxfreq=0,
                         blacklist=[], include_nonwords=0, wlnums='frq'):
    simple_n = float(simple_n)
    f = frq_db (sc, wlattr, wlnums)
    fref = frq_db (scref, wlattr, wlnums)
    size = sum(f)
    size_ref = sum(fref)
    p = 1.0 * size_ref / size
    attr = sc.get_attr(wlattr)
    attrref = scref.get_attr(wlattr)
    items = []
    if not include_nonwords: nwre = sc.get_conf('NONWORDRE')
    else: nwre = ''
    try: gen = attr.regexp2ids (wlpat.strip(), 0, nwre)
    except TypeError: gen = attr.regexp2ids (wlpat.strip(), 0)
    while not gen.end():
        if len(items) > 4 * wlmaxitems:
            items.sort(reverse=True)
            del items[wlmaxitems:]
        i = gen.next()
        w = attr.id2str(i)
        if f[i] < wlminfreq or (wlmaxfreq and f[i] > wlmaxfreq) \
                            or (words and w not in words) \
                            or (blacklist and w in blacklist):
            continue
        iref = attrref.str2id(w)
        fref_iref = (iref != -1 and fref[iref]) or 0
        if fref_iref == 0 or p * f[i] / fref[iref] > 1.0:
            rel = (f[i] * 1000000.0) / size
            relref = (fref_iref * 1000000.0) / size_ref
            score = (rel + simple_n) / (relref + simple_n)
            items.append ((score, rel, relref, f[i], fref_iref, w))
    items.sort(reverse=True)
    return items[:wlmaxitems]


def ngr_keywords (sc, scref, wlattr='word', wlminfreq=5, wlpat='.*',
                  wlmaxitems=100, simple_n=100, words=[], wlmaxfreq=0,
                  blacklist=[], include_nonwords=0, ngrams_n=2):
    import re
    ngr_data_filename = str(sc.get_conf('PATH') + wlattr + '.ngr')
    ngrref_data_filename = str(scref.get_conf('PATH') + wlattr + '.ngr')
    try: ngr = manatee.NGram(ngr_data_filename)
    except: raise MissingSubCorpFreqFile(ngr_data_filename)
    try: ngrref = manatee.NGram(ngrref_data_filename)
    except: raise MissingSubCorpFreqFile(ngrref_data_filename)
    if hasattr(sc, 'spath'):
        freqs = frq_db (sc, wlattr+'.ngr', id_range=ngr.size())
    if hasattr(scref, 'spath'):
        freqsref = frq_db (scref, wlattr+'.ngr', id_range=ngrref.size())
    size = sc.search_size()
    size_ref = scref.search_size()
    p = 1.0 * size_ref / size
    attr = sc.get_attr(wlattr)
    attrref = scref.get_attr(wlattr)
    items = []
    words, blacklist = set(words), set(blacklist)
    nwre = re.compile(sc.get_conf('NONWORDRE').replace('^[:alpha:]', '\d\W_'),
                      re.UNICODE)
    if wlpat and wlpat != '.*': wlpatre = re.compile(wlpat, re.UNICODE)
    else: wlpatre = None
    for i in xrange(ngr.size()):
        if hasattr(sc, 'spath'): freq = freqs[i]
        else: freq = ngr.freq(i)
        if freq < wlminfreq or (wlmaxfreq and freq > wlmaxfreq): continue
        if ngrams_n < ngr.min(i): continue # n-gram too big
        ngrwords = []
        lids = ngr.lids(i)
        whitelisted = False
        for n in xrange(ngr.max(i)): # "unzip" n-gram and filter
            word = attr.id2str(lids.next())
            if not include_nonwords and nwre.match(word): break
            if blacklist and word in blacklist: break
            if words and word in words: whitelisted = True
            ngrwords.append(word)
            if n+1 == ngrams_n: # check, compute score and append
                ngram = ' '.join(ngrwords)
                if wlpatre and not wlpatre.search(ngram): continue
                if words and not whitelisted and not ngram in words: continue
                if blacklist and ngram in blacklist: continue
                # locate ngram in scref
                refngram = ngrref.locate_id(map(attrref.str2id, ngrwords))
                if refngram == -1: reff = 0
                elif hasattr(scref, 'spath'): reff = freqsref[refngram]
                else: reff = ngrref.freq(refngram)
                if reff == 0 or p * freq / reff > 1.0:
                    rel = (freq * 1000000.0) / size
                    relref = (reff * 1000000.0) / size_ref
                    score = (rel + simple_n) / (relref + simple_n)
                    items.append ((score, rel, relref, freq, reff, ngram))
        if len(items) > 4 * wlmaxitems:
            items.sort(reverse=True)
            del items[wlmaxitems:]
    items.sort(reverse=True)
    return items[:wlmaxitems]


def ws_keywords (sc, scref, wlminfreq=10, wlmaxitems=100, simple_n=100,
                 ret_count=False, wlattr='ws_collocations'):
    import wmap
    wsbasekw = 'TERMBASE' if wlattr == 'ws_terms' else 'WSBASE'
    wmap.setEncoding('utf-8')
    basewsbase = os.path.basename(sc.get_conf(wsbasekw))
    ref_basewsbase = os.path.basename(scref.get_conf(wsbasekw))
    if not ref_basewsbase:
        raise RuntimeError('%s has no %s' % (scref.get_conf('NAME'), wsbasekw))
    freqs_file = subcorp_base_file(sc, basewsbase + '.hfrq')
    lex_file = subcorp_base_file(sc, basewsbase + '.hlex')
    ref_freqs_file = subcorp_base_file(scref, ref_basewsbase + '.hfrq')
    logfile = wlattr + '.build'

    if not os.path.isfile(freqs_file): # not computed
        raise MissingSubCorpFreqFile(sc)
    if not os.path.isfile(ref_freqs_file):
        raise MissingSubCorpFreqFile(scref)
    if os.path.isfile(subcorp_base_file(sc, logfile)):
        raise MissingSubCorpFreqFile(sc) # computation in progress
    if os.path.isfile(subcorp_base_file(scref, logfile)):
        raise MissingSubCorpFreqFile(scref)

    result_str = wmap.StrVector()
    wmap.extrms(freqs_file, ref_freqs_file, lex_file, result_str, wlmaxitems,
                simple_n)

    # find out seek and cnt -- firstly for sc ...
    ws1 = wmap.WMap (sc.get_conf(wsbasekw), 0, 0, 0, sc.get_conffile())
    size = sc.search_size()
    result_parsed = []
    try: wsstrip = int(sc.get_conf('WSSTRIP'))
    except ValueError: wsstrip = 0
    for item in result_str:
        w1, gramrel, w2, score = item.strip().split('\t')[:4]
        result_parsed.append ((ws1.coll2id(w1), ws1.str2id(gramrel),
                               ws1.coll2id(w2), w1, gramrel, w2, float(score)))
    result = {}
    for id1, id2, id3, w1, gramrel, w2, score in sorted (result_parsed):
        ws3 = ws_find_triple (ws1, id1, id2, id3)
        totalcount = ws3.size()
        if not ws3: continue
        freq = ws_subc_freq (ws3, sc)
        w2s = wsstrip and w2[:-wsstrip] or w2
        result[(-score, w1, gramrel, w2s)] = [freq, float(freq)*1000000/size,
                                              ws3.tell()]
    # ... and then for scref
    ws1 = wmap.WMap (scref.get_conf(wsbasekw), 0, 0, 0, scref.get_conffile())
    size = scref.search_size()
    result_parsed = []
    for item in result_str:
        w1, gramrel, w2, score = item.strip().split('\t')[:4]
        result_parsed.append ((ws1.coll2id(w1), ws1.str2id(gramrel),
                               ws1.coll2id(w2), w1, gramrel, w2, float(score)))
    for id1, id2, id3, w1, gramrel, w2, score in sorted (result_parsed):
        w2s = wsstrip and w2[:-wsstrip] or w2
        if not (-score, w1, gramrel, w2s) in result: continue
        ws3 = ws_find_triple (ws1, id1, id2, id3)
        if not ws3:
            result[(-score, w1, gramrel, w2s)].extend([0, 0, 0]); continue
        freq = ws_subc_freq (ws3, scref)
        result[(-score, w1, gramrel, w2s)].extend (
                                 [freq, float(freq)*1000000/size, ws3.tell()])
    result_list = [(-k[0], v[1], v[4], v[0], v[3],
                    '%s\t%s\t%s\t%d\t%d' % (k[1], k[2], k[3], v[2], v[5]))
                   for k, v in sorted (result.items())]
    if ret_count:
        return result_list, totalcount
    return result_list


def create_ws_db (corp, logfilename, wsbase):
    import subprocess
    basewsbase = os.path.basename(wsbase)
    freqs_file = subcorp_base_file(corp, basewsbase + '.hfrq')
    lex_file = subcorp_base_file(corp, basewsbase + '.hlex')
    logfile = open(logfilename, 'w')
    hashws_args = ['hashws', corp.get_conffile(), freqs_file, lex_file, wsbase]
    if hasattr(corp, 'spath'):
        hashws_args.append(corp.spath)
    ret = subprocess.call (hashws_args, stdout=logfile)
    if ret:
        open(logfilename + '.error', 'w').write(str(ret))
    logfile.close()
    os.rename (logfilename, logfilename + '.old')

def create_arf_db (corp, attrname, logfile=None, logstep=0.02):
    outfilename = subcorp_base_file (corp, attrname)
    if os.path.isfile (outfilename + '.arf') and os.path.isfile (outfilename
                                                                    + '.docf'):
        return
    path = ""
    if hasattr(corp, 'spath'):
        path = corp.spath
        sys.stderr.write ('trying find_same_subcorp_file: %s\n' % outfilename)
        same = corp.cm.find_same_subcorp_file (corp, attrname,
                                            ('arf','frq', 'docf', 'build.old'))
        sys.stderr.write ('find_same_subcorp_file: %s\n' % same)
        if same:
            same = same[:-4] + attrname
            complete = True
            for suff in ('.arf','.frq','.docf'):
                if not os.path.isfile(same + suff): complete = False
            if complete:
                from shutil import copyfile
                for suff in ('.arf','.frq','.docf'):
                    copyfile (same + suff, outfilename + suff)
                    try: copyfile (same + '.frq64', outfilename + '.frq64')
                    except: pass
                return

    if logfile:
        open(logfile, 'w').write('%d\n%s\n' % (os.getpid(), corp.search_size()))
        log = " 2>> " + logfile
    else:
        log = ""

    cmd = "mkstats '%s' '%s' %%s '%s' '%s'" % (corp.get_confpath(), attrname, path, log)
    import subprocess
    subprocess.call (cmd % "frq", shell=True)
    subprocess.call (cmd % "docf", shell=True)
    subprocess.call (cmd % "arf", shell=True)

    if logfile:
        os.rename (logfile, logfile + '.old')

def build_arf_db (corp, attrname):
    if '/' in attrname and attrname.endswith('.ngr'): # for genngr
        logfilename = attrname + '.build'
    else: # for others
        logfilename = subcorp_base_file (corp, attrname) + '.build'
    if os.path.isfile (logfilename):
        log = open(logfilename).read().split('\n')
        return log[0], log[-1].split('\r')[-1]
    sys.stderr.write ('build_arf_db:%s:%s (%s)\n' % (corp, attrname, logfilename)) 
    pid = os.fork()
    if pid == 0:
        import daemonize, subprocess
        # close only std{in,out,err}
        daemonize.createDaemon(maxfd=3)
        if hasattr(os, 'nice'):
            os.nice(10)
        if attrname == 'ws_terms':
            create_ws_db (corp, logfilename, corp.get_conf('TERMBASE'))
        elif attrname == 'ws_collocations':
            create_ws_db (corp, logfilename, corp.get_conf('WSBASE'))
        elif '/' in attrname and attrname.endswith('.ngr'):
            baseattrname = attrname[:-4].split('/')[-1]
            logfile = open(logfilename, 'a')
            logfile.write('%d\n%s\n0%%' % (os.getpid(), corp.search_size()))
            logfile.flush()
            subprocess.call(['genngr', corp.get_confpath(), baseattrname, '5',
                             attrname],
                           stdout=logfile, stderr=open(logfilename+'.err', 'w'))
            os.rename (logfilename, logfilename + '.old')
        else:
            create_arf_db (corp, attrname, logfilename)
        os._exit(0)
    else:
        return None

def build_arf_db_status (corp, attrname):
    logfilename = subcorp_base_file (corp, attrname) + '.build'
    if os.path.isfile (logfilename):
        log = open(logfilename).read().split('\n')
        return log[0], log[-1].split('\r')[-1]
    else:
        return 0, '100%'


def get_corp_info(corpname):
    import os, time
    from string import strip
    result = {
            'wposmap': False,
            'lposmap': False,
            'attrs': [],
            'structs': []
            }
    corp = manatee.Corpus(corpname)
    wposlist = corp.get_conf('WPOSLIST').split(',')
    lposlist = corp.get_conf('LPOSLIST').split(',')
    attrlist = corp.get_conf('ATTRLIST').split(',')
    structlist = corp.get_conf('STRUCTLIST').split(',')
    for item in attrlist:
        attr = corp.get_attr(item)
        result['attrs'].append((item, attr.id_range()))
    for item in structlist:
        struct = corp.get_struct(item)
        result['structs'].append((item, struct.size()))
    modtime = time.gmtime(os.stat(corp.get_conf('PATH')).st_mtime)
    sizes = dict([tuple(map(strip, i.split(' ')))
            for i in corp.get_sizes().split('\n') if i != ''])
    result.update({
            'name': corp.get_conf('NAME'),
            'lang': corp.get_conf('LANGUAGE'),
            'infohref': corp.get_conf('INFOHREF'),
            'encoding': corp.get_conf('ENCODING'),
            'tagsetdoc': corp.get_conf('TAGSETDOC'),
            'compiled': time.strftime("%m/%d/%Y %H:%M:%S", modtime),
            'tokens': sizes.get('tokencount', 0),
            'words': sizes.get('wordcount', 0),
            'pars': sizes.get('parcount', 0),
            'sents':sizes.get('sentcount', 0),
            'docs': sizes.get('doccount', 0),
            })
    if wposlist:
        result.update({'wposmap': zip(wposlist[1::2], wposlist[2::2])})
    if lposlist:
        result.update({'lposmap': zip(lposlist[1::2], lposlist[2::2])})
    return result


if __name__ == '__main__':
    import manatee
    if sys.argv[2:] and sys.argv[1] == '--create-arf-db':
        # ./corplib.py --create-arf-db CORPNAME [ATTRNAME]
        corp = manatee.Corpus(sys.argv[2])
        if sys.argv[3:]:
            attrnames = [sys.argv[3]]
        else:
            attrnames = corp.get_conf ('ATTRLIST').split(',')
        for attrn in attrnames:
            print 'Creating ARF database:', sys.argv[2], attrn
            create_arf_db (corp, attrn)
    else:
        cm = CorpusManager(subcpath=['/home/pary/corp/run/subcorp/GLOBAL'])
        os.environ['MANATEE_REGISTRY'] = '/home/pary/corp/registry'
        c=cm.get_Corpus('bnc:written')
        sc= cm.get_Corpus('bnc:academic')
